from abc import ABCMeta, abstractmethod
import os

class Repository:
	__metaclass__ = ABCMeta

	def __init__(self, repo):
		self.data = repo

	@abstractmethod
	def clone(self): pass

	@abstractmethod
	def pull(self): pass

	@abstractmethod
	def get_revision(self): pass


import util
from svnrepository import SVNRepository
from gitrepository import GitRepository

def create(repo):
	if util.isGitRepo(repo):
		return GitRepository(repo)
	else:
		return SVNRepository(repo)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import project

class ListDirs:
	def visit(self, host):
		for repo in host.tree:
			project.Project.currentProj += 1
			self.list(repo)

	def list(self, repo):
		print repo['local']


#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import os
import util
import project
import v_update

# Install is like update, but doesn't perform pulls.
class Install(v_update.Update):
	def pull(self, repo):
		return



#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import yaml
import os
import project
import dolly
import terminal
import sys

class Config:
	def __init__(self, path):
		self.data = self.__open(path)
		self.rootdir = dolly.Dolly.rootdir
	def __open(self, path):
		return yaml.load(open(path))

	def __normalizeTree(self, project):
		output = []
		for directory, repos in project.iteritems():
			for repository in repos:
				tag = ''
				branch = ''
				post_update = ''
				if 'repository' in repository:
					key = repository['name']
					remote = repository['repository']
					if 'tag' in repository:
						tag = repository['tag']
					if 'branch' in repository:
						branch = repository['branch']
					if 'post_update' in repository:
						post_update = repository['post_update']
				else:
					key = list(repository)[0]
					remote = repository[key]
				local = os.path.join(self.rootdir, directory, key)
				output.append({
					'local': local,
					'remote': remote,
					'name': key,
					'tag': tag,
					'branch': branch,
					'post_update': post_update
				})
		return output

	def parse(self, name):
		if not (name in self.data):
			terminal.error("Project " + name + " doesn't exist in config file")
			sys.exit(0)
		proj = self.data[name]
		tree = []
		includes = []
		description = ''
		post_update=None
		if 'tree' in proj:
			tree = self.__normalizeTree(proj['tree'])
		if 'includes' in proj:
			includes = proj['includes']
		if 'description' in proj:
			description = proj['description']
		if 'post_update' in proj:
			post_update = proj['post_update']

		return project.Project(name, description, tree, includes, self, post_update)

	def addRepo(self, remote, local, project):
		self.data[project]['tree'][os.path.relpath(os.path.abspath(os.path.join(local, '..')), self.rootdir)].append({os.path.basename(os.path.normpath(local)): remote})
		#self.data[project]['tree'][os.path.relpath(os.path.abspath(os.path.join(local, '..')), self.rootdir)][1] = remote

	def dump(self):
		print yaml.dump(self.data)




#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import subprocess
import project
import dolly
import terminal
import sys
import os

def isGitRepo(repo):
	return repo['remote'].endswith('.git')

def isSvnCheckout(repo):
    return os.path.exists(repo['local'] + '/.svn')

def isGitCheckout(repo):
    return os.path.exists(repo['local'] + '/.git')

def executeCommand(command, cwd=None):
	proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd, bufsize=0)
	output = ''
	if dolly.Dolly.verbose:
		print ''
		for line in iter(proc.stdout.readline, ''):
			sys.stdout.write(line)
			output += line
	stdout, stderr = proc.communicate()
	output += stdout
	returncode = proc.returncode
	if not returncode == 0:
		print ''
		terminal.error('Error while executing command "{0}"'.format(command))
		terminal.error(output)
		terminal.error(stderr)
		dolly.Dolly.warnings.append('Error while executing command "{0}" in "{1}"'.format(command, cwd))
	return {'returncode': returncode, 'stdout': output, 'stderr': stderr}


def printStatus(repo, count=True):
	if dolly.Dolly.verbose or count==False:
		if count:
			print ''
			print '({3}/{4}) [{0}] {1} => {2}'.format(repo['name'], repo['remote'], repo['local'], project.Project.currentProj, project.Project.projectCount)
		else:
			print '[{0}] {1} => {2}'.format(repo['name'], repo['remote'], repo['local'])
	else:
		terminal.clear_line()
		terminal.progress_bar(project.Project.projectCount, project.Project.currentProj, 30, repo['name'])

def checkRemote(repo):
	command = ''
	if isGitCheckout(repo):
		command = 'git remote -v'
	else:
		command = 'svn info '
	result = executeCommand(command, cwd=repo['local'])
	return (repo['remote'] in result['stdout']) or (result['returncode'] != 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 

import dolly
import sys

banner=r"""                       
     _     ,--.      |    |         
  _-(_)-   |   |,---.|    |    ,   .
`(___)     |   ||   ||    |    |   |
 // \\     `--' `---'`---'`---'`---|
                               `---'
"""

print banner

try:
	cmd = dolly.Dolly()
	cmd.run()
except KeyboardInterrupt:
	# Don't print a stacktrace.
	sys.exit(1)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import os
import util
import project
import terminal
import dolly

class Status:
	def visit(self, host):
		for repo in host.tree:
			project.Project.currentProj += 1
			util.printStatus(repo)
			self.status(repo)
	
	def status(self, repo):
		if not os.path.exists(repo['local']):
			dolly.Dolly.not_cloned.append(repo)
			return
		if not util.checkRemote(repo):
			error = "{0} has a different remote on disk than in config".format(repo['local'])
			terminal.error("\n" + error)
			dolly.Dolly.warnings.append(error)
		if util.isGitRepo(repo):
			self.statusGit(repo)
		else:
			self.statusSvn(repo)

	def statusGit(self, repo):
		if not util.isGitCheckout(repo):
			error = "{0} is a Git repo in config file but not a Git checkout on disk".format(repo['local'])
			terminal.error("\n" + error)
			dolly.Dolly.warnings.append(error)
			return
		result = util.executeCommand('git status -s', cwd=repo['local'])
		for line in result['stdout'].splitlines():
			change = os.path.join(repo['local'], ' '.join(line.split()[1:]))
			dolly.Dolly.changes.append({'repo': repo, 'change': change})
		result = util.executeCommand('git status -s -b', cwd=repo['local'])
		if result['stdout']:
			if 'ahead' in result['stdout']:
				dolly.Dolly.unpushed.append(repo)


	def statusSvn(self, repo):
		if not util.isSvnCheckout(repo):
			error = "{0} is a Svn repo in config file but not a Svn checkout on disk".format(repo['local'])
			terminal.error("\n" + error)
			dolly.Dolly.warnings.append(error)
			return
		result = util.executeCommand('svn status', cwd=repo['local'])
		for line in result['stdout'].splitlines():
			change = os.path.join(repo['local'], ' '.join(line.split()[1:]))
			dolly.Dolly.changes.append({'repo': repo, 'change': change})


#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import os
import sys
import util
import project
import terminal
import signal
from multiprocessing import Pool

import dolly
import repository

def init_worker():
	# Ignore SIGINTs in worker processes.
	signal.signal(signal.SIGINT, signal.SIG_IGN)

def process_repo(upd, repo):
	upd.update(repo)

class Update:
	def visit(self, host):
		pool = Pool(5, init_worker)

		def pr(repo):
			return pool.apply_async(process_repo, (self, repo))

		results = zip(host.tree, map(pr, host.tree))

		for r in results:
			repo, result = r
			project.Project.currentProj += 1
			util.printStatus(repo)
			# Workaround to Python issue 8296 where a SIGINT will
			# lock up the process when no wait time is given.
			result.wait(9999999)

		if host.post_update:
			util.executeCommand(host.post_update)
	
	def update(self, repo):
		r = repository.create(repo)
		if os.path.exists(repo['local']):
			self.pull(r)
		else:
			self.clone(r)

	def clone(self, repo):
		result = repo.clone()
		if result['returncode'] == 0:
			self.runPostUpdateCommand(repo)

	def pull(self, repo):
		if not util.checkRemote(repo.data):
			error = "{0} has a different remote on disk than in config".format(repo.data['local'])
			terminal.error("\n" + error)
			dolly.Dolly.warnings.append(error)

		revision = repo.get_revision()
		result = repo.pull()

		# Only run post-update commands when the repository's revision
		# was updated.
		if result['returncode'] == 0 and revision != repo.get_revision():
			self.runPostUpdateCommand(repo)



	def runPostUpdateCommand(self, repo):
		repo = repo.data
		if repo['post_update'] != '':
			util.executeCommand(repo['post_update'], cwd=repo['local'])



import dolly
import config
dolly.Dolly.rootdir="/home/lbentrup/dev"
p = config.Config("../config.yml")
print p.data
p.dump()
p.addRepo("git@github.com:foo/bar.git", "/home/lbentrup/dev/design/basic/foobar", "html")
print p.data
p.dump()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
class Project:
	projectCount = 0
	projects = []
	currentProj = 0

	def __init__(self, name, description, tree, includes, config, post_update=None):
		self.name = name
		self.description = description
		self.tree = tree
		self.includes = includes
		self.config = config
		
		self.post_update = post_update

	def accept(self, visitor):
		
		Project.projects.append(self.name)
		for include in self.includes:
			if not include in Project.projects:
				subproj = self.config.parse(include)
				subproj.accept(visitor)

		visitor.visit(self)


#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import argparse
import v_list
import v_count
import v_update
import v_status
import v_install
import v_list_dirs
import config
import os
import sys
import project
import terminal
import util
import pkg_resources

class Dolly:
	verbose = False
	rootdir = None
	not_cloned = []
	unpushed = []
	changes = []
	warnings = []

	def __init__(self):
		parser = argparse.ArgumentParser()
		parser.add_argument("command", help="Status/Update/List")
		parser.add_argument("project", help="The project to execute the command on", default="default", nargs="?")
		parser.add_argument("-v", "--verbose", action="store_true", help="Verbose Output")
		parser.add_argument("-r", "--rootdir", help="Clone repos relative to this path", default=os.path.expanduser('~/dev'))
		parser.add_argument("-c", "--config", help="Specify your dolly.yml")
		self.parser = parser
		self.args = parser.parse_args()
		Dolly.verbose = self.args.verbose
		Dolly.rootdir = self.args.rootdir
		self.startproject = None

	def run(self):
		if not self.args.command == 'list-dirs':
			self.banner()
		config_system = '/etc/dolly/dolly.yml'
		config_user = os.path.expanduser('~/.dolly.yml')
		if self.args.config:
			self.config=self.args.config
		elif os.path.exists(config_user):
			self.config = config_user
		elif os.path.exists(config_system):
			self.config = config_system
		else:
			terminal.error('No valid config file found')
			return
		parser = config.Config(self.config)
		self.startproject = parser.parse(self.args.project)
		self.countProjects()
		self.run_visitor()
		

	def run_visitor(self):
		command = self.args.command
		visitor = None
		if command in ['status', 'st']:
			visitor = v_status.Status()
		elif command in ['update', 'up']:
			visitor = v_update.Update()
		elif command in ['list', 'lst']:
			visitor = v_list.List()
		elif command in ['install', 'in']:
			visitor = v_install.Install()
		elif command in ['help', 'h']:
			self.parser.print_help()
			sys.exit(0)
		elif command in ['list-dirs']:
			visitor = v_list_dirs.ListDirs()
		elif command in ['version']:
			print "Dolly version: " + pkg_resources.require("Dolly")[0].version
			sys.exit(0)
		else:
			print 'Invalid command'
			sys.exit(0);
		self.startproject.accept(visitor)
		self.printSummary()

	def countProjects(self):
		counter = v_count.Count()
		self.startproject.accept(counter)
		project.Project.projects = []

	def printSummary(self):
		print '\n'
		if self.args.command in ['status', 'st']:
			terminal.blue('---SUMMARY---')
			print ''
			if len(Dolly.not_cloned) > 0:
				terminal.warning('The following repositories were not cloned')
				for repo in Dolly.not_cloned:
					util.printStatus(repo, False)
			print ''
			if len(Dolly.unpushed) > 0:
				terminal.warning('The following repositories contain unpushed commits')
				for repo in Dolly.unpushed:
					util.printStatus(repo, False)
			else:
				terminal.ok('No unpushed commits')
			print ''
			if len(Dolly.changes) > 0:
				terminal.warning('The following repositories contain uncomitted changes')
				for change in Dolly.changes:
					print '[{0}] {1}'.format(change['repo']['name'], change['change'])
			else:
				terminal.ok('No uncomitted changes')
			print ''
		if len(Dolly.warnings) > 0:
			terminal.warning('Some errors occured')
			for warning in Dolly.warnings:
				print warning
			sys.exit(1)
	def banner(self):
		banner=r"""                    
         _     ,--.      |    |         
      _-(_)-   |   |,---.|    |    ,   .
    `(___)     |   ||   ||    |    |   |
     // \\     `--' `---'`---'`---'`---|
                                   `---'
	"""

		print banner

def main():


	cmd = Dolly()
	cmd.run()

if __name__ == "__main__" :
	main()



from repository import Repository
import util

class SVNRepository(Repository):
	def __init__(self, repo):
		super(SVNRepository, self).__init__(repo)

	def clone(self):
		repo = self.data
		return util.executeCommand("svn checkout --config-option servers:global:store-plaintext-passwords=yes '{0}' '{1}'".format(repo['remote'], repo['local']))

	def pull(self):
		repo = self.data
		return util.executeCommand('svn update', cwd=repo['local'])

	def get_revision(self):
		repo = self.data
		result = util.executeCommand('svnversion', cwd=repo['local'])
		if result['returncode'] == 0:
			return result['stdout']
		else:
			return None



#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
import project

class List:
	def visit(self, host):
		for repo in host.tree:
			project.Project.currentProj += 1
			self.list(repo)

	def list(self, repo):
		print '({3}/{4}) [{0}] {1} => {2}'.format(repo['name'], repo['remote'], repo['local'], project.Project.currentProj, project.Project.projectCount)


from repository import Repository
import util

import os, errno

# Tries to create parent directories, avoiding race conditions.
#
# Git will do this automatically, but will fail when the directories are
# created in parallel.
def create_parent_directories(path):
	dir = os.path.dirname(path)
	if not os.path.exists(dir):
		try:
			os.makedirs(dir)
		except OSError, e:
			# We could check e.errno here for errno.EEXIST, but git will fail for all other errors anyways.
			return

class GitRepository(Repository):
	def __init__(self, repo):
		super(GitRepository, self).__init__(repo)

	def clone(self):
		repo = self.data
		create_parent_directories(repo['local'])
		branch = repo['branch']
		if repo['tag'] != '':
			branch = repo['tag']
		if branch != '':
			 result = util.executeCommand("git clone --branch '{2}' '{0}' '{1}'".format(
				 repo['remote'],
				 repo['local'],
				 branch
			 ))
		else:
			result = util.executeCommand("git clone '{0}' '{1}'".format(repo['remote'], repo['local']))
		return result

	def pull(self):
		repo = self.data
		if repo['tag'] != '':
			result = util.executeCommand("git checkout '{0}'".format(repo['tag']), cwd=repo['local'])
		else:
			result = util.executeCommand('git pull --ff-only', cwd=repo['local'])
		return result

	def get_revision(self):
		repo = self.data
		result = util.executeCommand('git rev-parse HEAD', cwd=repo['local'])
		if result['returncode'] == 0:
			return result['stdout']
		else:
			return None
